@@ -9,28 +9,58 @@ import (
 
 // Paths is specified by OpenAPI/Swagger standard version 3.
 // See https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#paths-object
-type Paths map[string]*PathItem
+type Paths struct {
+	Extensions map[string]interface{} `json:"-" yaml:"-"`
+
+	m map[string]*PathItem
+}
+
+// NewPathsWithCapacity builds a paths object of the given capacity.
+func NewPathsWithCapacity(cap int) *Paths {
+	return &Paths{m: make(map[string]*PathItem, cap)}
+}
+
+// NewPaths builds a paths object with path items in insertion order.
+func NewPaths(opts ...NewPathsOption) *Paths {
+	paths := NewPathsWithCapacity(len(opts))
+	for _, opt := range opts {
+		opt(paths)
+	}
+	return paths
+}
+
+// NewPathsOption describes options to NewPaths func
+type NewPathsOption func(*Paths)
+
+// WithPath adds a named path item
+func WithPath(path string, pathItem *PathItem) NewPathsOption {
+	return func(paths *Paths) {
+		if p := pathItem; p != nil && path != "" {
+			paths.Set(path, p)
+		}
+	}
+}
 
 // Validate returns an error if Paths does not comply with the OpenAPI spec.
-func (paths Paths) Validate(ctx context.Context, opts ...ValidationOption) error {
+func (paths *Paths) Validate(ctx context.Context, opts ...ValidationOption) error {
 	ctx = WithValidationOptions(ctx, opts...)
 
-	normalizedPaths := make(map[string]string, len(paths))
+	normalizedPaths := make(map[string]string, paths.Len())
 
-	keys := make([]string, 0, len(paths))
-	for key := range paths {
+	keys := make([]string, 0, paths.Len())
+	for key := range paths.Map() {
 		keys = append(keys, key)
 	}
 	sort.Strings(keys)
 	for _, path := range keys {
-		pathItem := paths[path]
+		pathItem := paths.Value(path)
 		if path == "" || path[0] != '/' {
 			return fmt.Errorf("path %q does not start with a forward slash (/)", path)
 		}
 
 		if pathItem == nil {
 			pathItem = &PathItem{}
-			paths[path] = pathItem
+			paths.Set(path, pathItem)
 		}
 
 		normalizedPath, _, varsInPath := normalizeTemplatedPath(path)
@@ -106,23 +136,23 @@ func (paths Paths) Validate(ctx context.Context, opts ...ValidationOption) error
 		return err
 	}
 
-	return nil
+	return validateExtensions(ctx, paths.Extensions)
 }
 
 // InMatchingOrder returns paths in the order they are matched against URLs.
 // See https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#paths-object
 // When matching URLs, concrete (non-templated) paths would be matched
 // before their templated counterparts.
-func (paths Paths) InMatchingOrder() []string {
+func (paths *Paths) InMatchingOrder() []string {
 	// NOTE: sorting by number of variables ASC then by descending lexicographical
 	// order seems to be a good heuristic.
-	if paths == nil {
+	if paths.Len() == 0 {
 		return nil
 	}
 
 	vars := make(map[int][]string)
 	max := 0
-	for path := range paths {
+	for path := range paths.Map() {
 		count := strings.Count(path, "}")
 		vars[count] = append(vars[count], path)
 		if count > max {
@@ -130,7 +160,7 @@ func (paths Paths) InMatchingOrder() []string {
 		}
 	}
 
-	ordered := make([]string, 0, len(paths))
+	ordered := make([]string, 0, paths.Len())
 	for c := 0; c <= max; c++ {
 		if ps, ok := vars[c]; ok {
 			sort.Sort(sort.Reverse(sort.StringSlice(ps)))
@@ -152,15 +182,15 @@ func (paths Paths) InMatchingOrder() []string {
 //	pathItem := path.Find("/person/{name}")
 //
 // would return the correct path item.
-func (paths Paths) Find(key string) *PathItem {
+func (paths *Paths) Find(key string) *PathItem {
 	// Try directly access the map
-	pathItem := paths[key]
+	pathItem := paths.Value(key)
 	if pathItem != nil {
 		return pathItem
 	}
 
 	normalizedPath, expected, _ := normalizeTemplatedPath(key)
-	for path, pathItem := range paths {
+	for path, pathItem := range paths.Map() {
 		pathNormalized, got, _ := normalizeTemplatedPath(path)
 		if got == expected && pathNormalized == normalizedPath {
 			return pathItem
@@ -169,9 +199,9 @@ func (paths Paths) Find(key string) *PathItem {
 	return nil
 }
 
-func (paths Paths) validateUniqueOperationIDs() error {
+func (paths *Paths) validateUniqueOperationIDs() error {
 	operationIDs := make(map[string]string)
-	for urlPath, pathItem := range paths {
+	for urlPath, pathItem := range paths.Map() {
 		if pathItem == nil {
 			continue
 		}
