@@ -4,88 +4,107 @@ import (
 	"context"
 	"encoding/json"
 	"errors"
-	"fmt"
 	"sort"
 	"strconv"
-
-	"github.com/go-openapi/jsonpointer"
 )
 
 // Responses is specified by OpenAPI/Swagger 3.0 standard.
 // See https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#responses-object
-type Responses map[string]*ResponseRef
+type Responses struct {
+	Extensions map[string]interface{} `json:"-" yaml:"-"`
 
-var _ jsonpointer.JSONPointable = (*Responses)(nil)
-
-func NewResponses() Responses {
-	r := make(Responses)
-	r["default"] = &ResponseRef{Value: NewResponse().WithDescription("")}
-	return r
+	m map[string]*ResponseRef
 }
 
-func (responses Responses) Default() *ResponseRef {
-	return responses["default"]
+// NewResponsesWithCapacity builds a responses object of the given capacity.
+func NewResponsesWithCapacity(cap int) *Responses {
+	return &Responses{m: make(map[string]*ResponseRef, cap)}
 }
 
-// Get returns a ResponseRef for the given status
+// NewResponses builds a responses object with response objects in insertion order.
+// Given no arguments, NewResponses returns a valid responses object containing a default match-all reponse.
+func NewResponses(opts ...NewResponsesOption) *Responses {
+	var responses *Responses
+	if n := len(opts); n != 0 {
+		responses = NewResponsesWithCapacity(n)
+	} else {
+		responses = &Responses{m: map[string]*ResponseRef{
+			"default": {Value: NewResponse().WithDescription("")},
+		}}
+	}
+	for _, opt := range opts {
+		opt(responses)
+	}
+	return responses
+}
+
+// NewResponsesOption describes options to NewResponses func
+type NewResponsesOption func(*Responses)
+
+// WithStatus adds a status code keyed ResponseRef
+func WithStatus(status int, responseRef *ResponseRef) NewResponsesOption {
+	return func(responses *Responses) {
+		if r := responseRef; r != nil {
+			code := strconv.FormatInt(int64(status), 10)
+			responses.Set(code, r)
+		}
+	}
+}
+
+// WithName adds a name-keyed Response
+func WithName(name string, response *Response) NewResponsesOption {
+	return func(responses *Responses) {
+		if r := response; r != nil && name != "" {
+			responses.Set(name, &ResponseRef{Value: r})
+		}
+	}
+}
+
+// Default returns the default response
+func (responses *Responses) Default() *ResponseRef {
+	return responses.Value("default")
+}
+
+// Status returns a ResponseRef for the given status
 // If an exact match isn't initially found a patterned field is checked using
 // the first digit to determine the range (eg: 201 to 2XX)
 // See https://spec.openapis.org/oas/v3.0.3#patterned-fields-0
-func (responses Responses) Get(status int) *ResponseRef {
+func (responses *Responses) Status(status int) *ResponseRef {
 	st := strconv.FormatInt(int64(status), 10)
-	if rref, ok := responses[st]; ok {
+	if rref := responses.Value(st); rref != nil {
 		return rref
 	}
-	st = string(st[0]) + "XX"
-	switch st {
-	case "1XX":
-		return responses["1XX"]
-	case "2XX":
-		return responses["2XX"]
-	case "3XX":
-		return responses["3XX"]
-	case "4XX":
-		return responses["4XX"]
-	case "5XX":
-		return responses["5XX"]
-	default:
-		return nil
-	}
-}
-
-// Validate returns an error if Responses does not comply with the OpenAPI spec.
-func (responses Responses) Validate(ctx context.Context, opts ...ValidationOption) error {
-	ctx = WithValidationOptions(ctx, opts...)
-
-	if len(responses) == 0 {
-		return errors.New("the responses object MUST contain at least one response code")
-	}
-
-	keys := make([]string, 0, len(responses))
-	for key := range responses {
-		keys = append(keys, key)
-	}
-	sort.Strings(keys)
-	for _, key := range keys {
-		v := responses[key]
-		if err := v.Validate(ctx); err != nil {
-			return err
+	if 99 < status && status < 600 {
+		st = string(st[0]) + "XX"
+		switch st {
+		case "1XX", "2XX", "3XX", "4XX", "5XX":
+			return responses.Value(st)
 		}
 	}
 	return nil
 }
 
-// JSONLookup implements https://pkg.go.dev/github.com/go-openapi/jsonpointer#JSONPointable
-func (responses Responses) JSONLookup(token string) (interface{}, error) {
-	ref, ok := responses[token]
-	if !ok {
-		return nil, fmt.Errorf("invalid token reference: %q", token)
+// Validate returns an error if Responses does not comply with the OpenAPI spec.
+func (responses *Responses) Validate(ctx context.Context, opts ...ValidationOption) error {
+	ctx = WithValidationOptions(ctx, opts...)
+
+	if responses.Len() == 0 {
+		return errors.New("the responses object MUST contain at least one response code")
 	}
 
-	if ref != nil && ref.Ref != "" {
-		return &Ref{Ref: ref.Ref}, nil
+	keys := make([]string, 0, responses.Len())
+	for key := range responses.Map() {
+		keys = append(keys, key)
 	}
-	return ref.Value, nil
+	sort.Strings(keys)
+	for _, key := range keys {
+		v := responses.Value(key)
+		if err := v.Validate(ctx); err != nil {
+			return err
+		}
+	}
+
+	return validateExtensions(ctx, responses.Extensions)
 }
 
 // Response is specified by OpenAPI/Swagger 3.0 standard.
@@ -149,13 +168,16 @@ func (response *Response) UnmarshalJSON(data []byte) error {
 	type ResponseBis Response
 	var x ResponseBis
 	if err := json.Unmarshal(data, &x); err != nil {
-		return err
+		return unmarshalError(err)
 	}
 	_ = json.Unmarshal(data, &x.Extensions)
 	delete(x.Extensions, "description")
 	delete(x.Extensions, "headers")
 	delete(x.Extensions, "content")
 	delete(x.Extensions, "links")
+	if len(x.Extensions) == 0 {
+		x.Extensions = nil
+	}
 	*response = Response(x)
 	return nil
 }
