@@ -13,6 +13,7 @@ import (
 	"sort"
 	"strconv"
 	"strings"
+	"sync"
 	"unicode/utf16"
 
 	"github.com/go-openapi/jsonpointer"
@@ -47,28 +48,10 @@ var (
 	ErrSchemaInputNaN = errors.New("floating point NaN is not allowed")
 	// ErrSchemaInputInf may be returned when validating a number
 	ErrSchemaInputInf = errors.New("floating point Inf is not allowed")
+
+	compiledPatterns sync.Map
 )
 
-// Float64Ptr is a helper for defining OpenAPI schemas.
-func Float64Ptr(value float64) *float64 {
-	return &value
-}
-
-// BoolPtr is a helper for defining OpenAPI schemas.
-func BoolPtr(value bool) *bool {
-	return &value
-}
-
-// Int64Ptr is a helper for defining OpenAPI schemas.
-func Int64Ptr(value int64) *int64 {
-	return &value
-}
-
-// Uint64Ptr is a helper for defining OpenAPI schemas.
-func Uint64Ptr(value uint64) *uint64 {
-	return &value
-}
-
 // NewSchemaRef simply builds a SchemaRef
 func NewSchemaRef(ref string, value *Schema) *SchemaRef {
 	return &SchemaRef{
@@ -77,28 +60,11 @@ func NewSchemaRef(ref string, value *Schema) *SchemaRef {
 	}
 }
 
-type Schemas map[string]*SchemaRef
-
-var _ jsonpointer.JSONPointable = (*Schemas)(nil)
-
-// JSONLookup implements github.com/go-openapi/jsonpointer#JSONPointable
-func (s Schemas) JSONLookup(token string) (interface{}, error) {
-	ref, ok := s[token]
-	if ref == nil || ok == false {
-		return nil, fmt.Errorf("object has no field %q", token)
-	}
-
-	if ref.Ref != "" {
-		return &Ref{Ref: ref.Ref}, nil
-	}
-	return ref.Value, nil
-}
-
 type SchemaRefs []*SchemaRef
 
 var _ jsonpointer.JSONPointable = (*SchemaRefs)(nil)
 
-// JSONLookup implements github.com/go-openapi/jsonpointer#JSONPointable
+// JSONLookup implements https://pkg.go.dev/github.com/go-openapi/jsonpointer#JSONPointable
 func (s SchemaRefs) JSONLookup(token string) (interface{}, error) {
 	i, err := strconv.ParseUint(token, 10, 64)
 	if err != nil {
@@ -154,10 +120,9 @@ type Schema struct {
 	MultipleOf *float64 `json:"multipleOf,omitempty" yaml:"multipleOf,omitempty"`
 
 	// String
-	MinLength       uint64  `json:"minLength,omitempty" yaml:"minLength,omitempty"`
-	MaxLength       *uint64 `json:"maxLength,omitempty" yaml:"maxLength,omitempty"`
-	Pattern         string  `json:"pattern,omitempty" yaml:"pattern,omitempty"`
-	compiledPattern *regexp.Regexp
+	MinLength uint64  `json:"minLength,omitempty" yaml:"minLength,omitempty"`
+	MaxLength *uint64 `json:"maxLength,omitempty" yaml:"maxLength,omitempty"`
+	Pattern   string  `json:"pattern,omitempty" yaml:"pattern,omitempty"`
 
 	// Array
 	MinItems uint64     `json:"minItems,omitempty" yaml:"minItems,omitempty"`
@@ -196,7 +161,7 @@ func (addProps AdditionalProperties) MarshalJSON() ([]byte, error) {
 func (addProps *AdditionalProperties) UnmarshalJSON(data []byte) error {
 	var x interface{}
 	if err := json.Unmarshal(data, &x); err != nil {
-		return err
+		return unmarshalError(err)
 	}
 	switch y := x.(type) {
 	case nil:
@@ -360,7 +325,7 @@ func (schema *Schema) UnmarshalJSON(data []byte) error {
 	type SchemaBis Schema
 	var x SchemaBis
 	if err := json.Unmarshal(data, &x); err != nil {
-		return err
+		return unmarshalError(err)
 	}
 	_ = json.Unmarshal(data, &x.Extensions)
 
@@ -413,6 +378,10 @@ func (schema *Schema) UnmarshalJSON(data []byte) error {
 	delete(x.Extensions, "additionalProperties")
 	delete(x.Extensions, "discriminator")
 
+	if len(x.Extensions) == 0 {
+		x.Extensions = nil
+	}
+
 	*schema = Schema(x)
 
 	if schema.Format == "date" {
@@ -424,7 +393,7 @@ func (schema *Schema) UnmarshalJSON(data []byte) error {
 	return nil
 }
 
-// JSONLookup implements github.com/go-openapi/jsonpointer#JSONPointable
+// JSONLookup implements https://pkg.go.dev/github.com/go-openapi/jsonpointer#JSONPointable
 func (schema Schema) JSONLookup(token string) (interface{}, error) {
 	switch token {
 	case "additionalProperties":
@@ -712,7 +681,6 @@ func (schema *Schema) WithMaxLengthDecodedBase64(i int64) *Schema {
 
 func (schema *Schema) WithPattern(pattern string) *Schema {
 	schema.Pattern = pattern
-	schema.compiledPattern = nil
 	return schema
 }
 
@@ -809,35 +777,35 @@ func (schema *Schema) IsEmpty() bool {
 		schema.MinProps != 0 || schema.MaxProps != nil {
 		return false
 	}
-	if n := schema.Not; n != nil && !n.Value.IsEmpty() {
+	if n := schema.Not; n != nil && n.Value != nil && !n.Value.IsEmpty() {
 		return false
 	}
-	if ap := schema.AdditionalProperties.Schema; ap != nil && !ap.Value.IsEmpty() {
+	if ap := schema.AdditionalProperties.Schema; ap != nil && ap.Value != nil && !ap.Value.IsEmpty() {
 		return false
 	}
 	if apa := schema.AdditionalProperties.Has; apa != nil && !*apa {
 		return false
 	}
-	if items := schema.Items; items != nil && !items.Value.IsEmpty() {
+	if items := schema.Items; items != nil && items.Value != nil && !items.Value.IsEmpty() {
 		return false
 	}
 	for _, s := range schema.Properties {
-		if !s.Value.IsEmpty() {
+		if ss := s.Value; ss != nil && !ss.IsEmpty() {
 			return false
 		}
 	}
 	for _, s := range schema.OneOf {
-		if !s.Value.IsEmpty() {
+		if ss := s.Value; ss != nil && !ss.IsEmpty() {
 			return false
 		}
 	}
 	for _, s := range schema.AnyOf {
-		if !s.Value.IsEmpty() {
+		if ss := s.Value; ss != nil && !ss.IsEmpty() {
 			return false
 		}
 	}
 	for _, s := range schema.AllOf {
-		if !s.Value.IsEmpty() {
+		if ss := s.Value; ss != nil && !ss.IsEmpty() {
 			return false
 		}
 	}
@@ -847,60 +815,70 @@ func (schema *Schema) IsEmpty() bool {
 // Validate returns an error if Schema does not comply with the OpenAPI spec.
 func (schema *Schema) Validate(ctx context.Context, opts ...ValidationOption) error {
 	ctx = WithValidationOptions(ctx, opts...)
-	return schema.validate(ctx, []*Schema{})
+	_, err := schema.validate(ctx, []*Schema{})
+	return err
 }
 
-func (schema *Schema) validate(ctx context.Context, stack []*Schema) error {
+// returns the updated stack and an error if Schema does not comply with the OpenAPI spec.
+func (schema *Schema) validate(ctx context.Context, stack []*Schema) ([]*Schema, error) {
 	validationOpts := getValidationOptions(ctx)
 
 	for _, existing := range stack {
 		if existing == schema {
-			return nil
+			return stack, nil
 		}
 	}
 	stack = append(stack, schema)
 
 	if schema.ReadOnly && schema.WriteOnly {
-		return errors.New("a property MUST NOT be marked as both readOnly and writeOnly being true")
+		return stack, errors.New("a property MUST NOT be marked as both readOnly and writeOnly being true")
 	}
 
 	for _, item := range schema.OneOf {
 		v := item.Value
 		if v == nil {
-			return foundUnresolvedRef(item.Ref)
+			return stack, foundUnresolvedRef(item.Ref)
 		}
-		if err := v.validate(ctx, stack); err != nil {
-			return err
+
+		var err error
+		if stack, err = v.validate(ctx, stack); err != nil {
+			return stack, err
 		}
 	}
 
 	for _, item := range schema.AnyOf {
 		v := item.Value
 		if v == nil {
-			return foundUnresolvedRef(item.Ref)
+			return stack, foundUnresolvedRef(item.Ref)
 		}
-		if err := v.validate(ctx, stack); err != nil {
-			return err
+
+		var err error
+		if stack, err = v.validate(ctx, stack); err != nil {
+			return stack, err
 		}
 	}
 
 	for _, item := range schema.AllOf {
 		v := item.Value
 		if v == nil {
-			return foundUnresolvedRef(item.Ref)
+			return stack, foundUnresolvedRef(item.Ref)
 		}
-		if err := v.validate(ctx, stack); err != nil {
-			return err
+
+		var err error
+		if stack, err = v.validate(ctx, stack); err != nil {
+			return stack, err
 		}
 	}
 
 	if ref := schema.Not; ref != nil {
 		v := ref.Value
 		if v == nil {
-			return foundUnresolvedRef(ref.Ref)
+			return stack, foundUnresolvedRef(ref.Ref)
 		}
-		if err := v.validate(ctx, stack); err != nil {
-			return err
+
+		var err error
+		if stack, err = v.validate(ctx, stack); err != nil {
+			return stack, err
 		}
 	}
 
@@ -914,7 +892,7 @@ func (schema *Schema) validate(ctx context.Context, stack []*Schema) error {
 			case "float", "double":
 			default:
 				if validationOpts.schemaFormatValidationEnabled {
-					return unsupportedFormat(format)
+					return stack, unsupportedFormat(format)
 				}
 			}
 		}
@@ -924,7 +902,7 @@ func (schema *Schema) validate(ctx context.Context, stack []*Schema) error {
 			case "int32", "int64":
 			default:
 				if validationOpts.schemaFormatValidationEnabled {
-					return unsupportedFormat(format)
+					return stack, unsupportedFormat(format)
 				}
 			}
 		}
@@ -946,31 +924,33 @@ func (schema *Schema) validate(ctx context.Context, stack []*Schema) error {
 			default:
 				// Try to check for custom defined formats
 				if _, ok := SchemaStringFormats[format]; !ok && validationOpts.schemaFormatValidationEnabled {
-					return unsupportedFormat(format)
+					return stack, unsupportedFormat(format)
 				}
 			}
 		}
-		if schema.Pattern != "" && !validationOpts.schemaPatternValidationDisabled {
-			if err := schema.compilePattern(); err != nil {
-				return err
+		if !validationOpts.schemaPatternValidationDisabled && schema.Pattern != "" {
+			if _, err := schema.compilePattern(); err != nil {
+				return stack, err
 			}
 		}
 	case TypeArray:
 		if schema.Items == nil {
-			return errors.New("when schema type is 'array', schema 'items' must be non-null")
+			return stack, errors.New("when schema type is 'array', schema 'items' must be non-null")
 		}
 	case TypeObject:
 	default:
-		return fmt.Errorf("unsupported 'type' value %q", schemaType)
+		return stack, fmt.Errorf("unsupported 'type' value %q", schemaType)
 	}
 
 	if ref := schema.Items; ref != nil {
 		v := ref.Value
 		if v == nil {
-			return foundUnresolvedRef(ref.Ref)
+			return stack, foundUnresolvedRef(ref.Ref)
 		}
-		if err := v.validate(ctx, stack); err != nil {
-			return err
+
+		var err error
+		if stack, err = v.validate(ctx, stack); err != nil {
+			return stack, err
 		}
 	}
 
@@ -983,45 +963,49 @@ func (schema *Schema) validate(ctx context.Context, stack []*Schema) error {
 		ref := schema.Properties[name]
 		v := ref.Value
 		if v == nil {
-			return foundUnresolvedRef(ref.Ref)
+			return stack, foundUnresolvedRef(ref.Ref)
 		}
-		if err := v.validate(ctx, stack); err != nil {
-			return err
+
+		var err error
+		if stack, err = v.validate(ctx, stack); err != nil {
+			return stack, err
 		}
 	}
 
 	if schema.AdditionalProperties.Has != nil && schema.AdditionalProperties.Schema != nil {
-		return errors.New("additionalProperties are set to both boolean and schema")
+		return stack, errors.New("additionalProperties are set to both boolean and schema")
 	}
 	if ref := schema.AdditionalProperties.Schema; ref != nil {
 		v := ref.Value
 		if v == nil {
-			return foundUnresolvedRef(ref.Ref)
+			return stack, foundUnresolvedRef(ref.Ref)
 		}
-		if err := v.validate(ctx, stack); err != nil {
-			return err
+
+		var err error
+		if stack, err = v.validate(ctx, stack); err != nil {
+			return stack, err
 		}
 	}
 
 	if v := schema.ExternalDocs; v != nil {
 		if err := v.Validate(ctx); err != nil {
-			return fmt.Errorf("invalid external docs: %w", err)
+			return stack, fmt.Errorf("invalid external docs: %w", err)
 		}
 	}
 
 	if v := schema.Default; v != nil && !validationOpts.schemaDefaultsValidationDisabled {
 		if err := schema.VisitJSON(v); err != nil {
-			return fmt.Errorf("invalid default: %w", err)
+			return stack, fmt.Errorf("invalid default: %w", err)
 		}
 	}
 
 	if x := schema.Example; x != nil && !validationOpts.examplesValidationDisabled {
 		if err := validateExampleValue(ctx, x, schema); err != nil {
-			return fmt.Errorf("invalid example: %w", err)
+			return stack, fmt.Errorf("invalid example: %w", err)
 		}
 	}
 
-	return validateExtensions(ctx, schema.Extensions)
+	return stack, validateExtensions(ctx, schema.Extensions)
 }
 
 func (schema *Schema) IsMatching(value interface{}) bool {
@@ -1062,7 +1046,11 @@ func (schema *Schema) VisitJSON(value interface{}, opts ...SchemaValidationOptio
 func (schema *Schema) visitJSON(settings *schemaValidationSettings, value interface{}) (err error) {
 	switch value := value.(type) {
 	case nil:
-		return schema.visitJSONNull(settings)
+		// Don't use VisitJSONNull, as we still want to reach 'visitXOFOperations', since
+		// those could allow for a nullable value even though this one doesn't
+		if schema.Nullable {
+			return
+		}
 	case float64:
 		if math.IsNaN(value) {
 			return ErrSchemaInputNaN
@@ -1073,13 +1061,28 @@ func (schema *Schema) visitJSON(settings *schemaValidationSettings, value interf
 	}
 
 	if schema.IsEmpty() {
+		switch value.(type) {
+		case nil:
+			return schema.visitJSONNull(settings)
+		default:
+			return
+		}
+	}
+
+	if err = schema.visitNotOperation(settings, value); err != nil {
 		return
 	}
-	if err = schema.visitSetOperations(settings, value); err != nil {
+	var run bool
+	if err, run = schema.visitXOFOperations(settings, value); err != nil || !run {
+		return
+	}
+	if err = schema.visitEnumOperation(settings, value); err != nil {
 		return
 	}
 
 	switch value := value.(type) {
+	case nil:
+		return schema.visitJSONNull(settings)
 	case bool:
 		return schema.visitJSONBoolean(settings, value)
 	case json.Number:
@@ -1140,7 +1143,7 @@ func (schema *Schema) visitJSON(settings *schemaValidationSettings, value interf
 	}
 }
 
-func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, value interface{}) (err error) {
+func (schema *Schema) visitEnumOperation(settings *schemaValidationSettings, value interface{}) (err error) {
 	if enum := schema.Enum; len(enum) != 0 {
 		for _, v := range enum {
 			switch c := value.(type) {
@@ -1152,6 +1155,10 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 				if v == f {
 					return
 				}
+			case int64:
+				if v == float64(c) {
+					return
+				}
 			default:
 				if reflect.DeepEqual(v, value) {
 					return
@@ -1170,7 +1177,10 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 			customizeMessageError: settings.customizeMessageError,
 		}
 	}
+	return
+}
 
+func (schema *Schema) visitNotOperation(settings *schemaValidationSettings, value interface{}) (err error) {
 	if ref := schema.Not; ref != nil {
 		v := ref.Value
 		if v == nil {
@@ -1188,7 +1198,13 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 			}
 		}
 	}
+	return
+}
 
+// If the XOF operations pass successfully, abort further run of validation, as they will already be satisfied (unless the schema
+// itself is badly specified
+func (schema *Schema) visitXOFOperations(settings *schemaValidationSettings, value interface{}) (err error, run bool) {
+	var visitedOneOf, visitedAnyOf, visitedAllOf bool
 	if v := schema.OneOf; len(v) > 0 {
 		var discriminatorRef string
 		if schema.Discriminator != nil {
@@ -1200,7 +1216,7 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 						Schema:      schema,
 						SchemaField: "discriminator",
 						Reason:      fmt.Sprintf("input does not contain the discriminator property %q", pn),
-					}
+					}, false
 				}
 
 				discriminatorValString, okcheck := discriminatorVal.(string)
@@ -1210,7 +1226,7 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 						Schema:      schema,
 						SchemaField: "discriminator",
 						Reason:      fmt.Sprintf("value of discriminator property %q is not a string", pn),
-					}
+					}, false
 				}
 
 				if discriminatorRef, okcheck = schema.Discriminator.Mapping[discriminatorValString]; len(schema.Discriminator.Mapping) > 0 && !okcheck {
@@ -1219,7 +1235,7 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 						Schema:      schema,
 						SchemaField: "discriminator",
 						Reason:      fmt.Sprintf("discriminator property %q has invalid value", pn),
-					}
+					}, false
 				}
 			}
 		}
@@ -1233,7 +1249,7 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 		for idx, item := range v {
 			v := item.Value
 			if v == nil {
-				return foundUnresolvedRef(item.Ref)
+				return foundUnresolvedRef(item.Ref), false
 			}
 
 			if discriminatorRef != "" && discriminatorRef != item.Ref {
@@ -1256,7 +1272,7 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 
 		if ok != 1 {
 			if settings.failfast {
-				return errSchema
+				return errSchema, false
 			}
 			e := &SchemaError{
 				Value:                 value,
@@ -1272,13 +1288,14 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 				e.Reason = `value doesn't match any schema from "oneOf"`
 			}
 
-			return e
+			return e, false
 		}
 
 		// run again to inject default value that defined in matched oneOf schema
 		if settings.asreq || settings.asrep {
 			_ = v[matchedOneOfIndices[0]].Value.visitJSON(settings, value)
 		}
+		visitedOneOf = true
 	}
 
 	if v := schema.AnyOf; len(v) > 0 {
@@ -1290,7 +1307,7 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 		for idx, item := range v {
 			v := item.Value
 			if v == nil {
-				return foundUnresolvedRef(item.Ref)
+				return foundUnresolvedRef(item.Ref), false
 			}
 			// make a deep copy to protect origin value from being injected default value that defined in mismatched anyOf schema
 			if settings.asreq || settings.asrep {
@@ -1304,7 +1321,7 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 		}
 		if !ok {
 			if settings.failfast {
-				return errSchema
+				return errSchema, false
 			}
 			return &SchemaError{
 				Value:                 value,
@@ -1312,20 +1329,21 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 				SchemaField:           "anyOf",
 				Reason:                `doesn't match any schema from "anyOf"`,
 				customizeMessageError: settings.customizeMessageError,
-			}
+			}, false
 		}
 
 		_ = v[matchedAnyOfIdx].Value.visitJSON(settings, value)
+		visitedAnyOf = true
 	}
 
 	for _, item := range schema.AllOf {
 		v := item.Value
 		if v == nil {
-			return foundUnresolvedRef(item.Ref)
+			return foundUnresolvedRef(item.Ref), false
 		}
 		if err := v.visitJSON(settings, value); err != nil {
 			if settings.failfast {
-				return errSchema
+				return errSchema, false
 			}
 			return &SchemaError{
 				Value:                 value,
@@ -1334,9 +1352,12 @@ func (schema *Schema) visitSetOperations(settings *schemaValidationSettings, val
 				Reason:                `doesn't match all schemas from "allOf"`,
 				Origin:                err,
 				customizeMessageError: settings.customizeMessageError,
-			}
+			}, false
 		}
+		visitedAllOf = true
 	}
+
+	run = !(visitedOneOf || visitedAnyOf || visitedAllOf)
 	return
 }
 
@@ -1389,7 +1410,7 @@ func (schema *Schema) visitJSONNumber(settings *schemaValidationSettings, value
 				Value:                 value,
 				Schema:                schema,
 				SchemaField:           "type",
-				Reason:                fmt.Sprintf("value must be an integer"),
+				Reason:                "value must be an integer",
 				customizeMessageError: settings.customizeMessageError,
 			}
 			if !settings.multiError {
@@ -1596,28 +1617,32 @@ func (schema *Schema) visitJSONString(settings *schemaValidationSettings, value
 	}
 
 	// "pattern"
-	if schema.Pattern != "" && schema.compiledPattern == nil && !settings.patternValidationDisabled {
-		var err error
-		if err = schema.compilePattern(); err != nil {
+	if !settings.patternValidationDisabled && schema.Pattern != "" {
+		cpiface, _ := compiledPatterns.Load(schema.Pattern)
+		cp, _ := cpiface.(*regexp.Regexp)
+		if cp == nil {
+			var err error
+			if cp, err = schema.compilePattern(); err != nil {
+				if !settings.multiError {
+					return err
+				}
+				me = append(me, err)
+			}
+		}
+		if !cp.MatchString(value) {
+			err := &SchemaError{
+				Value:                 value,
+				Schema:                schema,
+				SchemaField:           "pattern",
+				Reason:                fmt.Sprintf(`string doesn't match the regular expression "%s"`, schema.Pattern),
+				customizeMessageError: settings.customizeMessageError,
+			}
 			if !settings.multiError {
 				return err
 			}
 			me = append(me, err)
 		}
 	}
-	if cp := schema.compiledPattern; cp != nil && !cp.MatchString(value) {
-		err := &SchemaError{
-			Value:                 value,
-			Schema:                schema,
-			SchemaField:           "pattern",
-			Reason:                fmt.Sprintf(`string doesn't match the regular expression "%s"`, schema.Pattern),
-			customizeMessageError: settings.customizeMessageError,
-		}
-		if !settings.multiError {
-			return err
-		}
-		me = append(me, err)
-	}
 
 	// "format"
 	var formatStrErr string
@@ -1969,18 +1994,6 @@ func (schema *Schema) expectedType(settings *schemaValidationSettings, value int
 	}
 }
 
-func (schema *Schema) compilePattern() (err error) {
-	if schema.compiledPattern, err = regexp.Compile(schema.Pattern); err != nil {
-		return &SchemaError{
-			Schema:      schema,
-			SchemaField: "pattern",
-			Origin:      err,
-			Reason:      fmt.Sprintf("cannot compile pattern %q: %v", schema.Pattern, err),
-		}
-	}
-	return nil
-}
-
 // SchemaError is an error that occurs during schema validation.
 type SchemaError struct {
 	// Value is the value that failed validation.
@@ -2093,8 +2106,8 @@ func isSliceOfUniqueItems(xs []interface{}) bool {
 	s := len(xs)
 	m := make(map[string]struct{}, s)
 	for _, x := range xs {
-		// The input slice is coverted from a JSON string, there shall
-		// have no error when covert it back.
+		// The input slice is converted from a JSON string, there shall
+		// have no error when convert it back.
 		key, _ := json.Marshal(&x)
 		m[string(key)] = struct{}{}
 	}
